searchNodes=[{"doc":"What is ElvenGard Currently, all independent developers wishing to create a Multiplayer game have already asked themselves the question of how to make the server part easily while being &quot;solid&quot; (minimum crash, latency, ...). Indeed, for the client part, there are many tools very well designed to realize it (Unity3D, Unreal Engine,...) but for the server part, each game being different there are currently very few solutions to do this work. This is the goal of this ambitious project: make a toolkit to group together different functionalities present in any MMORPG (network part, quests, movements, objects in game, instances, etc...) to prepare bases for the developer so that he doesn't have to dwell on this part which is often tedious. What is ElvenGard.Network ElvenGard.Network is a dedicated toolkit designed to streamline and enhance the network aspect of game server development. Built on top of the &quot;ranch&quot; Erlang library, this powerful toolkit provides developers with a robust foundation for creating multiplayer game servers with ease. Key Features: Specialized Network Handling: ElvenGard.Network is solely focused on handling the complexities of networking in multiplayer game servers. It offers a high-performance solution optimized for managing player connections, communication, and data exchange. Efficient Protocol Management: The toolkit simplifies the implementation of network protocols, allowing developers to define and manage packet structures efficiently. This streamlined approach ensures smooth and reliable communication between the game client and server. Packet Processing Made Easy: ElvenGard.Network streamlines packet processing, making it straightforward to handle incoming and outgoing data. Developers can effortlessly manage packet reception, interpretation, and response, thereby reducing the burden of low-level network management. Custom Serialization/Deserialization: With support for custom data types, ElvenGard.Network enables seamless serialization and deserialization of data. This feature ensures compatibility and efficient data exchange between different components of the game server. Minimized Latency and Crashes: By leveraging the &quot;ranch&quot; library, ElvenGard.Network benefits from Erlang's reliability and scalability. This helps minimize latency issues and reduces the risk of crashes, providing a stable networking foundation for multiplayer games. With ElvenGard.Network handling the intricate network tasks, game developers can focus on creating captivating gameplay, dynamic worlds, and engaging player experiences. By leveraging this specialized toolkit, the journey from conceptualizing a multiplayer game to deploying a robust server becomes significantly smoother and more efficient. Installation The package can be installed by adding elvengard_network to your list of dependencies in mix.exs : def deps do [ { :elvengard_network , &quot;~&gt; 0.1.1&quot; } ] end The docs can be found at https://hexdocs.pm/elvengard_network . /!\\ This toolkit is currently not production ready ! Projects using ElvenGard MinecraftEx : Located in the examples folder of the repository, this is the beginning of a Minecraft server emulator AvantHeim : Created by the same developer as ElvenGard, this is a NosTale server emulator Contributing I'm currently developing this project. Any review or PR is welcome. Also, feel free to fork the repository and contribute.","ref":"ElvenGard.Network.html","title":"ElvenGard.Network","type":"module"},{"doc":"Manage a socket. This module provides functionality for managing a socket in the network protocol. A socket is a connection between the server and a client. It maintains various socket fields, such as the socket ID, socket assigns, transport information, and the packet network encoder used for sending data. Socket fields :id : The unique string ID of the socket. :assigns : A map of socket assigns, which can be used to store custom data associated with the socket. The default value is %{} . :transport : The Ranch transport used for the socket. :transport_pid : The PID (Process ID) of the socket's transport process. :remaining : The remaining bytes after receiving and packet deserialization. :encoder : The ElvenGard.Network.NetworkCodec module used to encode packets in the send/2 function.","ref":"ElvenGard.Network.Socket.html","title":"ElvenGard.Network.Socket","type":"module"},{"doc":"","ref":"ElvenGard.Network.Socket.html#assign/2","title":"ElvenGard.Network.Socket.assign/2","type":"function"},{"doc":"Adds key value pairs to socket assigns. A single key value pair may be passed, a keyword list or map of assigns may be provided to be merged into existing socket assigns. Examples iex&gt; assign ( socket , :name , &quot;ElvenGard&quot; ) iex&gt; socket . assigns . name == &quot;ElvenGard&quot; true iex&gt; assign ( socket , name : &quot;ElvenGard&quot; , logo : &quot;🌸&quot; ) iex&gt; socket . assigns . name == &quot;ElvenGard&quot; true iex&gt; socket . assigns . logo == &quot;🌸&quot; true","ref":"ElvenGard.Network.Socket.html#assign/3","title":"ElvenGard.Network.Socket.assign/3","type":"function"},{"doc":"Create a new socket structure. This function initializes a new socket with the given transport_pid , transport , and encoder module.","ref":"ElvenGard.Network.Socket.html#new/3","title":"ElvenGard.Network.Socket.new/3","type":"function"},{"doc":"Send a packet to the client. This function sends a packet to the client through the socket's transport. If the socket's encoder is set to :unset , the data is sent as is. Otherwise, the encoder module is used to serialize the data before sending it. Examples iex&gt; ElvenGard.Network.Socket . send ( socket , % LoginResponse { status : 200 , message : &quot;Welcome!&quot; } ) :ok","ref":"ElvenGard.Network.Socket.html#send/2","title":"ElvenGard.Network.Socket.send/2","type":"function"},{"doc":"","ref":"ElvenGard.Network.Socket.html#t:t/0","title":"ElvenGard.Network.Socket.t/0","type":"type"},{"doc":"Define a behaviour for custom types (packet parsing). This module defines a behaviour that allows users to create custom packet parsing types. By implementing the callbacks defined in this module, users can define how a specific field type is decoded and encoded when parsing and generating packets. To implement a custom type, you need to define the decode/2 and encode/2 callbacks. The decode/2 function takes a raw binary as input and decodes it into a structured value of the custom type. The encode/2 function takes a value of the custom type and encodes it into a binary representation. Note that the ElvenGard.Network.PacketSerializer module uses the callbacks in this behaviour when encoding and decoding packet fields. For each field defined in a packet serializer, the corresponding type's encode/2 or decode/2 function will be called to parse the binary data. Example Here's an example of defining a custom type for decoding and encoding a 16-bit little-endian integer: defmodule MyLittleEndianIntType do use ElvenGard.Network.Type @impl ElvenGard.Network.Type def decode ( raw , _opts ) do &lt;&lt; value :: little - integer - size ( 16 ) , rest :: binary &gt;&gt; = raw { value , rest } end @impl ElvenGard.Network.Type def encode ( value , _opts ) do &lt;&lt; value :: little - integer - size ( 16 ) &gt;&gt; end end In the above example, we defined a custom type module MyLittleEndianIntType that decodes and encodes a 16-bit little-endian integer. Then, when defining a packet serializer using ElvenGard.Network.PacketSerializer , you can use this custom type to define a field in the packet structure.","ref":"ElvenGard.Network.Type.html","title":"ElvenGard.Network.Type","type":"behaviour"},{"doc":"Decode a raw bitstring into a structured value of the custom type. Arguments: raw : The raw bitstring to be decoded. opts : A keyword list of options that can be used during decoding (optional). The function should return a tuple with the decoded value and the remaining unparsed bitstring.","ref":"ElvenGard.Network.Type.html#c:decode/2","title":"ElvenGard.Network.Type.decode/2","type":"callback"},{"doc":"Encode a value of the custom type into binary data. Arguments: value : The value of the custom type to be encoded. opts : A keyword list of options that can be used during encoding (optional). The function should return the encoded binary data as an iodata or a bitstring.","ref":"ElvenGard.Network.Type.html#c:encode/2","title":"ElvenGard.Network.Type.encode/2","type":"callback"},{"doc":"Define a custom behaviour for a View (packet sent from server to client). This module defines a behaviour that allows users to create custom packet views to be sent from the server to the client. By implementing the callback defined in this module, users can define how a specific packet view is rendered and converted into binary data for transmission. To implement a custom view, you need to define the render/2 callback. The render/2 function takes the name of the packet view and a map of parameters as input and returns the binary data to be sent to the client or a structure that will be serialized by ElvenGard.Network.PacketSerializer . Example Here's an example of defining a custom view for rendering a login response packet: defmodule MyApp.LoginViews do use ElvenGard.Network.View alias MyApp.Server.LoginPackets.LoginResponse @impl ElvenGard.Network.View def render ( :login_response , %{ status : status , message : message } ) do # Encode the `json` field before generating the LoginResponse struct json = Poison . encode! ( message ) % LoginResponse { status : status , json : json } end end In the above example, we defined a custom views module MyApp.LoginViews that implements the render/2 callback for rendering a login response packet.","ref":"ElvenGard.Network.View.html","title":"ElvenGard.Network.View","type":"behaviour"},{"doc":"Build a packet to send to the client. Arguments: name : A unique identifier packet view (string or atom). params : A map or keyword list of parameters to be used when rendering the view. The function should return an iodata or a structure to be sent to the client.","ref":"ElvenGard.Network.View.html#c:render/2","title":"ElvenGard.Network.View.render/2","type":"callback"},{"doc":"Wrapper on top of Ranch listeners . This module provides a wrapper around the Ranch library to define network endpoints. Endpoints are crucial for managing incoming connections and handling network traffic efficiently. For in-depth information on how to use and configure network endpoints, please refer to the Endpoint documentation .","ref":"ElvenGard.Network.Endpoint.html","title":"ElvenGard.Network.Endpoint","type":"behaviour"},{"doc":"Called just before starting the ranch listener","ref":"ElvenGard.Network.Endpoint.html#c:handle_start/1","title":"ElvenGard.Network.Endpoint.handle_start/1","type":"callback"},{"doc":"Wrapper on top of Ranch protocols . This module defines a protocol behavior to handle incoming connections in the ElvenGard.Network library. It provides callbacks for initializing, handling incoming messages, and handling connection termination. This protocol behavior serves as a wrapper around Ranch protocols, providing a structured way to implement connection handling within ElvenGard.Network. For detailed information on implementing and using network protocols with ElvenGard.Network, please refer to the Endpoint Protocol guide .","ref":"ElvenGard.Network.Endpoint.Protocol.html","title":"ElvenGard.Network.Endpoint.Protocol","type":"behaviour"},{"doc":"Callback called after the socket connection is closed and before the GenServer shutdown.","ref":"ElvenGard.Network.Endpoint.Protocol.html#c:handle_halt/2","title":"ElvenGard.Network.Endpoint.Protocol.handle_halt/2","type":"callback"},{"doc":"Callback called just before entering the GenServer loop. This callback is invoked when a new connection is established and before the GenServer loop starts processing messages. For the return values, see GenServer.init/1","ref":"ElvenGard.Network.Endpoint.Protocol.html#c:handle_init/1","title":"ElvenGard.Network.Endpoint.Protocol.handle_init/1","type":"callback"},{"doc":"Callback called just after receiving a message. This callback is invoked whenever a message is received on the connection. It should return one of the following: :ignore : the message received will not be decoded or processed by the protocol. It will just be ignored {:ignore, new_socket} : same as :ignore but also modifies the socket {:ok, new_socket} : classic loop - decode the packet and process it {:stop, reason, new_socket} : stop the GenServer/Protocol and disconnect the client","ref":"ElvenGard.Network.Endpoint.Protocol.html#c:handle_message/2","title":"ElvenGard.Network.Endpoint.Protocol.handle_message/2","type":"callback"},{"doc":"Behavior for Defining Packet Encoding and Decoding Implementations of this behavior are responsible for parsing incoming raw binary data into structured packets and encoding structured packets into binary data for transmission over the network. For more information on how to implement this behavior and use packet encoding and decoding, please refer to the NetworkCodec guide .","ref":"ElvenGard.Network.NetworkCodec.html","title":"ElvenGard.Network.NetworkCodec","type":"behaviour"},{"doc":"Decodes a packet from raw binary data. This callback function is responsible for decoding a packet from the provided raw binary data. The decoded packet should be returned as a struct containing the parsed fields. Examples raw_packet = &lt;&lt; 1 :: 8 , 123 :: 8 , &quot;some data&quot; :: binary &gt;&gt; # Example raw packet data % MessageStruct { id : 123 , data : &quot;some data&quot; } = MyNetworkCodec . decode ( raw_packet , socket )","ref":"ElvenGard.Network.NetworkCodec.html#c:decode/2","title":"ElvenGard.Network.NetworkCodec.decode/2","type":"callback"},{"doc":"Encodes a packet for network transmission. This callback function is responsible for encoding a structured packet or raw binary data into the binary format suitable for transmission over the network. The resulting encoded data, in the form of iodata() , can be directly sent over the network. Examples packet = % MessageStruct { id : 1 , data : &quot;Hello&quot; } &lt;&lt; header :: 8 , id :: 8 , data :: binary &gt;&gt; = MyNetworkCodec . encode ( packet , socket )","ref":"ElvenGard.Network.NetworkCodec.html#c:encode/2","title":"ElvenGard.Network.NetworkCodec.encode/2","type":"callback"},{"doc":"Identifies and extracts the first packet from raw binary data. This function searches for the first packet within the raw binary data and returns it along with the remaining binary data. The extracted packet will be passed to the decode/2 callback for further processing. Examples raw = &lt;&lt; packet1 :: binary , remaining :: binary &gt;&gt; { ^ packet1 , ^ remaining } = MyNetworkCodec . next ( raw , socket )","ref":"ElvenGard.Network.NetworkCodec.html#c:next/2","title":"ElvenGard.Network.NetworkCodec.next/2","type":"callback"},{"doc":"Provides a behavior for handling incoming packets. This module defines the callback handle_packet/2 , which must be implemented by modules using this behavior to process incoming packets. For detailed usage information, please refer to the packet handler guide .","ref":"ElvenGard.Network.PacketHandler.html","title":"ElvenGard.Network.PacketHandler","type":"behaviour"},{"doc":"This callback function is called after a packet is received and decoded by the Endpoint. It is responsible for processing the packet data and returning one of the following tuples: {:cont, socket} : Indicates that packet processing is complete, and the client connection should continue. {:halt, socket} : Indicates that packet processing is complete, and the client connection should be terminated. Parameters packet : The packet data as a struct. socket : The socket representing the client connection.","ref":"ElvenGard.Network.PacketHandler.html#c:handle_packet/2","title":"ElvenGard.Network.PacketHandler.handle_packet/2","type":"callback"},{"doc":"Packet Serializer DSL for defining packet structures. This module provides a DSL (Domain-Specific Language) for defining packet serializers for both received and sent packets in a network protocol. It enables users to create structured packets with specified fields and decode binary data into packet structs. To learn more about the available macros and how to define packet serializers, refer to the Packet Serializer DSL guide . Packet Macros The defpacket macros ( defpacket/1 , defpacket/2 , and defpacket/3 ) allow users to define packet structures. They require a packet ID and an alias (which is the name of the generated packet structure). The guard (with the when keyword) and do-block (for defining fields) are optional. Users can specify guards in packet macros to conditionally match packets based on a condition (often using socket assigns). Packet Structure, Serialization, and Deserialization The defpacket macros can generate a packet structure, a deserialize/3 function for deserializing binary data into the packet structure, and a serialize/1 function for generating the serialized binary representation of the packet. The deserialize/3 function should be used for decoding received packets, and the serialize/1 function should be used for generating packets to be sent over the network. Field Macros The field macros are used to define fields within a packet: field/2 : Define a field with a name and type. field/3 : Define a field with a name, type, and decoding options. Decorators The following decorators can be used to specify serialization and deserialization properties: @serializable true : Marks the packet as serializable for sending over the network. @deserializable true : Marks the packet as deserializable for receiving from the network.","ref":"ElvenGard.Network.PacketSerializer.html","title":"ElvenGard.Network.PacketSerializer","type":"module"},{"doc":"","ref":"ElvenGard.Network.PacketSerializer.html#defpacket/1","title":"ElvenGard.Network.PacketSerializer.defpacket/1","type":"macro"},{"doc":"","ref":"ElvenGard.Network.PacketSerializer.html#defpacket/2","title":"ElvenGard.Network.PacketSerializer.defpacket/2","type":"macro"},{"doc":"","ref":"ElvenGard.Network.PacketSerializer.html#defpacket/3","title":"ElvenGard.Network.PacketSerializer.defpacket/3","type":"macro"},{"doc":"","ref":"ElvenGard.Network.PacketSerializer.html#field/3","title":"ElvenGard.Network.PacketSerializer.field/3","type":"macro"},{"doc":"","ref":"ElvenGard.Network.PacketSerializer.html#import_packets/1","title":"ElvenGard.Network.PacketSerializer.import_packets/1","type":"macro"},{"doc":"Exception raised when a View is not found","ref":"ElvenGard.Network.UnknownViewError.html","title":"ElvenGard.Network.UnknownViewError","type":"exception"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Add .formatter.exs file to release for import_deps","ref":"changelog.html#0-1-1","title":"Changelog - 0.1.1","type":"extras"},{"doc":"Pre-release, this is not a production ready release !","ref":"changelog.html#0-1-0","title":"Changelog - 0.1.0","type":"extras"},{"doc":"","ref":"readme.html","title":"Overview","type":"extras"},{"doc":"Currently, all independent developers wishing to create a Multiplayer game have already asked themselves the question of how to make the server part easily while being &quot;solid&quot; (minimum crash, latency, ...). Indeed, for the client part, there are many tools very well designed to realize it (Unity3D, Unreal Engine,...) but for the server part, each game being different there are currently very few solutions to do this work. This is the goal of this ambitious project: make a toolkit to group together different functionalities present in any MMORPG (network part, quests, movements, objects in game, instances, etc...) to prepare bases for the developer so that he doesn't have to dwell on this part which is often tedious.","ref":"readme.html#what-is-elvengard","title":"Overview - What is ElvenGard","type":"extras"},{"doc":"ElvenGard.Network is a dedicated toolkit designed to streamline and enhance the network aspect of game server development. Built on top of the &quot;ranch&quot; Erlang library, this powerful toolkit provides developers with a robust foundation for creating multiplayer game servers with ease. Key Features: Specialized Network Handling: ElvenGard.Network is solely focused on handling the complexities of networking in multiplayer game servers. It offers a high-performance solution optimized for managing player connections, communication, and data exchange. Efficient Protocol Management: The toolkit simplifies the implementation of network protocols, allowing developers to define and manage packet structures efficiently. This streamlined approach ensures smooth and reliable communication between the game client and server. Packet Processing Made Easy: ElvenGard.Network streamlines packet processing, making it straightforward to handle incoming and outgoing data. Developers can effortlessly manage packet reception, interpretation, and response, thereby reducing the burden of low-level network management. Custom Serialization/Deserialization: With support for custom data types, ElvenGard.Network enables seamless serialization and deserialization of data. This feature ensures compatibility and efficient data exchange between different components of the game server. Minimized Latency and Crashes: By leveraging the &quot;ranch&quot; library, ElvenGard.Network benefits from Erlang's reliability and scalability. This helps minimize latency issues and reduces the risk of crashes, providing a stable networking foundation for multiplayer games. With ElvenGard.Network handling the intricate network tasks, game developers can focus on creating captivating gameplay, dynamic worlds, and engaging player experiences. By leveraging this specialized toolkit, the journey from conceptualizing a multiplayer game to deploying a robust server becomes significantly smoother and more efficient.","ref":"readme.html#what-is-elvengard-network","title":"Overview - What is ElvenGard.Network","type":"extras"},{"doc":"The package can be installed by adding elvengard_network to your list of dependencies in mix.exs : def deps do [ { :elvengard_network , &quot;~&gt; 0.1.1&quot; } ] end The docs can be found at https://hexdocs.pm/elvengard_network . /!\\ This toolkit is currently not production ready !","ref":"readme.html#installation","title":"Overview - Installation","type":"extras"},{"doc":"MinecraftEx : Located in the examples folder of the repository, this is the beginning of a Minecraft server emulator AvantHeim : Created by the same developer as ElvenGard, this is a NosTale server emulator","ref":"readme.html#projects-using-elvengard","title":"Overview - Projects using ElvenGard","type":"extras"},{"doc":"I'm currently developing this project. Any review or PR is welcome. Also, feel free to fork the repository and contribute.","ref":"readme.html#contributing","title":"Overview - Contributing","type":"extras"},{"doc":"Before we start coding, we need to define our network protocol, i.e. what our packets are going to look like, how they are going to be encoded, decoded and so on. For this guide, we're going to use a text-based protocol , meaning that all commands exchanged are text that can be read by a human. We'll be using the String module for parsing and tools such as Netcat as a client.","ref":"network_protocol.html","title":"Network Protocol","type":"extras"},{"doc":"In this section we'll look at the different packets and interactions between client and server. Ping A client can send a PingRequest : Packet ID Field Name Field Type Notes PING no fields When the server receives this packet, it must respond with a PongResponse : Packet ID Field Name Field Type Notes PONG time DateTime Time at which the server received the ping requests Login A client can send a LoginRequest : Packet ID Field Name Field Type Notes LOGIN username String password String Depending on whether the credentials are correct or not, the server can respond in 2 ways, LoginFailed : Packet ID Field Name Field Type Notes FAIL reason String The reason why the user cannot log in or LoginSucceed : Packet ID Field Name Field Type Notes SUCCESS world_info WorldInfo Host and port of our game server","ref":"network_protocol.html#packets","title":"Network Protocol - Packets","type":"extras"},{"doc":"Now that we know what our packets will look like, we need to define how we're going to exchange them on the network. As Elvengard.Network is currently based on Ranch , it only supports the TCP protocol . And, as we use a text protocol, we'll just separate our fields with a space and our different packets with a line breaks ( \\n ). Here's an example of what each packet might look like once encoded: PingRequest: PING\\n PongResponse: PONG 2023-08-15 20:45:07.068297Z\\n LoginRequest: LOGIN admin password\\n LoginFailed: FAIL Bad credentials\\n LoginSucceed: SUCCESS 127.0.0.1:5000\\n","ref":"network_protocol.html#network-encoding-decoding","title":"Network Protocol - Network encoding/decoding","type":"extras"},{"doc":"Now that you know how your client and server will communicate, you can create and configure your project. That's what we'll look at in the next section.","ref":"network_protocol.html#summary","title":"Network Protocol - Summary","type":"extras"},{"doc":"This guide is an introduction to ElvenGard.Network , the Network toolkit for a MMO Game Server toolkit written in Elixir. The purpose of this toolkit is to provide a set of modules and processes to simplify the creation of a game server. It's therefore simple to use and flexible in order to allow you to concentrate on the implementation of your features. In this guide, we're going to learn some basics about ElvenGard.Network. If you want to see the code from this guide, you can view it at elvengard_network/examples/login_server on GitHub .","ref":"getting_started.html","title":"Getting Started","type":"extras"},{"doc":"To start off with, we'll generate a new Elixir application by running this command: mix new login_server -- sup The --sup option ensures that this application has a supervision tree , which we'll need for ElvenGard a little later on. To add ElvenGard.Network to this application, just add an entry to your mix.exs: defp deps do [ { :elvengard_network , &quot;~&gt; 0.1.1&quot; } ] end To install this dependency, we will run this command: mix deps . get Then, we need to add a bit of configuration in config/config.exs import Config config :login_server , LoginServer.Endpoint , listener_name : :login_server , transport : :ranch_tcp , transport_opts : [ ip : &quot;127.0.0.1&quot; , port : 3000 ] , protocol : LoginServer.Endpoint.Protocol config :login_server , LoginServer.Endpoint.Protocol , packet_handler : LoginServer.Endpoint.PacketHandler , network_codec : LoginServer.Endpoint.NetworkCodec This configuration example simply tells our endpoint to listen on port 3000 of the local address and defines the protocol, network encoder and packet handler to be used. For more details on the configuration of each module, please refer to Endpoint and Protocol guides and Ranch documentation .","ref":"getting_started.html#adding-elvengard-network-to-an-application","title":"Getting Started - Adding ElvenGard.Network to an application","type":"extras"},{"doc":"At the end of this section, you must have setup your app to use Elvengard.Network. Now that we have everything installed, let's create our first Endpoint and get up and running.","ref":"getting_started.html#summary","title":"Getting Started - Summary","type":"extras"},{"doc":"The first thing to do is define a module using ElvenGard.Network.Endpoint . An Endpoint is simply a Ranch listener .","ref":"endpoint.html","title":"Endpoint","type":"extras"},{"doc":"Configuration is performed via the config/config.exs file as you did previously. config :login_server , LoginServer.Endpoint , listener_name : :login_server , transport : :ranch_tcp , transport_opts : [ ip : &quot;127.0.0.1&quot; , port : 3000 ] , protocol : LoginServer.Endpoint.Protocol Here's what each configuration option does: listener_name: :login_server : specifies a unique name for the listener. This is used by Ranch to manage the listener process. transport: :ranch_tcp : specifies the transport protocol to use. In this case, it's Ranch's TCP transport. transport_opts: [ip: &quot;127.0.0.1&quot;, port: 3000] : provides options for configuring the transport. In this case, it specifies the IP address and port on which the server will listen for incoming connections. protocol: LoginServer.Endpoint.Protocol : sets the protocol module that will handle client connections and communication.","ref":"endpoint.html#configuration","title":"Endpoint - Configuration","type":"extras"},{"doc":"Here's a basic example you can use for all your projects: # file: lib/login_server/endpoint.ex defmodule LoginServer.Endpoint do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.Endpoint &quot;&quot;&quot; use ElvenGard.Network.Endpoint , otp_app : :login_server require Logger ## Callbacks @impl true def handle_start ( config ) do host = get_in ( config , [ :transport_opts , :socket_opts , :ip ] ) port = get_in ( config , [ :transport_opts , :socket_opts , :port ] ) Logger . info ( &quot;LoginServer started on \#{ :inet . ntoa ( host ) } : \#{ port } &quot; ) end end As you can see, creating an endpoint is very simple, you just need to specify the otp app you used in the config and you're done. It is also possible to define the ElvenGard.Network.Endpoint.handle_start/1 callback. This allows you to, for example, display various information relating to the startup. It receives the endpoint configs as a parameter and must always return :ok .","ref":"endpoint.html#creating-an-endpoint","title":"Endpoint - Creating an Endpoint","type":"extras"},{"doc":"Once you've created and configured your endpoint, all you have to do is place it in your application's supervision tree. # file: lib/login_server/application.ex ... def start ( _type , _args ) do children = [ LoginServer.Endpoint ] ...","ref":"endpoint.html#add-to-supervision-tree","title":"Endpoint - Add to supervision tree","type":"extras"},{"doc":"At the end of this part, you should have a working Endpoint listening on port 3000 (see config file). Now it's time to create a Protocol to receive our first packets.","ref":"endpoint.html#summary","title":"Endpoint - Summary","type":"extras"},{"doc":"In this section, we will learn how to use ElvenGard.Network.Endpoint.Protocol . An Protocol is wrapper around Ranch protocols .","ref":"protocol.html","title":"Protocol","type":"extras"},{"doc":"Configuration is performed via the config/config.exs file as you did previously. config :login_server , LoginServer.Endpoint.Protocol , packet_handler : LoginServer.Endpoint.PacketHandler , network_codec : LoginServer.Endpoint.NetworkCodec Here's the explanation for these options: packet_handler: LoginServer.PacketHandler : specifies the module responsible for handling packets received from clients. network_codec: LoginServer.NetworkCodec : Specifies the module responsible for encoding and decoding packets for communication between clients and the server.","ref":"protocol.html#configuration","title":"Protocol - Configuration","type":"extras"},{"doc":"For this part, we're going to create a fairly simple protocol that will just display our packets. # file: lib/login_server/endpoint/protocol.ex defmodule LoginServer.Endpoint.Protocol do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.Endpoint.Protocol &quot;&quot;&quot; use ElvenGard.Network.Endpoint.Protocol require Logger alias ElvenGard.Network.Socket ## Callbacks @impl true def handle_init ( % Socket { } = socket ) do Logger . info ( &quot;New connection: \#{ socket . id } &quot; ) % Socket { transport : transport , transport_pid : transport_pid } = socket :ok = transport . setopts ( transport_pid , packet : :line , reuseaddr : true ) { :ok , socket } end @impl true def handle_message ( message , % Socket { } = socket ) do Logger . debug ( &quot;New message from \#{ socket . id } : \#{ inspect ( message ) } &quot; ) :ignore end @impl true def handle_halt ( reason , % Socket { } = socket ) do Logger . info ( &quot; \#{ socket . id } is now disconnected (reason: \#{ inspect ( reason ) } )&quot; ) { :ok , socket } end end Once again, creating a Protocol is fairly straightforward. This example defines 3 callbacks : handle_init/1 : called when a client connects, it is mainly used to set socket options or call ElvenGard.Network.Socket.assign/2 to init assigns. handle_message/2 : called when we receive a packet from a client, we can either ignore it by returning :ignore , or choose to decode it and then handle it by returning :ok . handle_halt/2 : called when a client disconnects. NOTE : you may notice that we define the packet: :line option in handle_init/1 . We use this option because we want to use a line break as a separator for our packets. This works because, according to our network protocol , we use a text protocol where each packet is separated by a \\n . However, for a binary protocol, you may need to use packet: :raw or other options. For more information on available options, see :inet.setopts/2 .","ref":"protocol.html#create-a-protocol","title":"Protocol - Create a Protocol","type":"extras"},{"doc":"If you run your application in its current state, you'll see that it's possible to connect to our server and send it messages, and that these are displayed by our application. Since we're using a text-based protocol, you can use Netcat as client for example. Now that packets can be received, they need to be decoded and processed. This is what we'll see next.","ref":"protocol.html#summary","title":"Protocol - Summary","type":"extras"},{"doc":"In this section, we will learn how to use ElvenGard.Network.Type . We're going to create all the types and subpackets needed to serialize and deserialize our packets according to our network protocol .","ref":"types_and_subpackets.html","title":"Types and Sub Packets","type":"extras"},{"doc":"So, according to our network protocol, we must define theses types: Field Type Encoded or Decoded Used by String Both LoginRequest, LoginFailed DateTime Encoded PongResponse Let's start with the first one: # file: lib/login_server/types/string_type.ex defmodule LoginServer.Types.StringType do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.Types.StringType &quot;&quot;&quot; use ElvenGard.Network.Type @type t :: String . t ( ) ## Behaviour impls @impl true @spec decode ( binary ( ) , Keyword . t ( ) ) :: { t ( ) , binary ( ) } def decode ( data , _opts ) when is_binary ( data ) do case String . split ( data , &quot; &quot; , parts : 2 ) do [ string ] -&gt; { string , &quot;&quot; } [ string , rest ] -&gt; { string , rest } end end @impl true @spec encode ( t ( ) , Keyword . t ( ) ) :: binary ( ) def encode ( data , _opts ) when is_binary ( data ) do data end end As you can see, we just need to use ElvenGard.Network.Type and define 2 callbacks: decode/2 : takes the binary to be decoded and options. This callback must return a tuple in the form {type_decoded, rest_of_binary_not_decoded} . encode/2 : takes the type to encode and options. This callback must return an iodata representation of our type. NOTE : Typespecs and guards are not mandatory, but are a good practice. # file: lib/login_server/types/date_time_type.ex defmodule LoginServer.Types.DateTimeType do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.Types.DateTimeType &quot;&quot;&quot; use ElvenGard.Network.Type @type t :: DateTime . t ( ) ## Behaviour impls @impl true def decode ( _data , _opts ) , do : raise ( &quot;unimplemented&quot; ) @impl true @spec encode ( t ( ) , Keyword . t ( ) ) :: binary ( ) def encode ( data , _opts ) when is_struct ( data , DateTime ) do DateTime . to_string ( data ) end end For this second type it's the same, the only difference is that we don't need to need to deserialize a DateTimeType , so we can just ignore the function and raise if someone is trying to use it.","ref":"types_and_subpackets.html#types","title":"Types and Sub Packets - Types","type":"extras"},{"doc":"A sub-packet is simply a type that uses other types. Unlike a packet, it has no ID, as it will be used by other packets. Let's look at a simple example with the WorldInfo : SubPacket Name Encoded or Decoded Used by WorldInfo Encoded LoginSucceed # file: lib/login_server/sub_packets/world_info.ex defmodule LoginServer.SubPackets.WorldInfo do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.SubPackets.WorldInfo &quot;&quot;&quot; use ElvenGard.Network.Type alias __MODULE__ alias LoginServer.Types . { IntegerType , StringType } @enforce_keys [ :host , :port ] defstruct [ :host , :port ] @type t :: % WorldInfo { host : StringType . t ( ) , port : IntegerType . t ( ) } ## Behaviour impls @impl true def decode ( _data , _opts ) , do : raise ( &quot;unimplemented&quot; ) @impl true @spec encode ( t ( ) , Keyword . t ( ) ) :: iolist ( ) def encode ( data , opts ) when is_struct ( data , WorldInfo ) do separator = Keyword . fetch! ( opts , :sep ) [ StringType . encode ( data . host ) , StringType . encode ( separator ) , IntegerType . encode ( data . port ) ] end end As you can see, this type is represented in Elixir by a structure with 2 mandatory fields: host and port . They are each represented by the StringType and IntegerType types, with which they will be encoded. NOTE : this sub-packet has another special feature: the separator used by the fields can be configured through options. We'll see how to use it in the next sections. As the IntegerType type has not been created yet, let's create it defining only the encode/2 function. # file: lib/login_server/types/integer_type.ex defmodule LoginServer.Types.IntegerType do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.Types.IntegerType &quot;&quot;&quot; use ElvenGard.Network.Type @type t :: integer ( ) ## Behaviour impls @impl true def decode ( _data , _opts ) , do : raise ( &quot;unimplemented&quot; ) @impl true @spec encode ( t ( ) , Keyword . t ( ) ) :: binary ( ) def encode ( data , _opts ) when is_integer ( data ) do Integer . to_string ( data ) end end","ref":"types_and_subpackets.html#sub-packets","title":"Types and Sub Packets - Sub Packets","type":"extras"},{"doc":"In this guide, we've seen how to create customizable types and sub-packets. Now we'll look at how to use them.","ref":"types_and_subpackets.html#summary","title":"Types and Sub Packets - Summary","type":"extras"},{"doc":"In this section, we will learn how to use ElvenGard.Network.PacketSerializer . We'll use the previously created types to transform our network protocol into Elixir structures.","ref":"packet_definitions.html","title":"Packet Definitions","type":"extras"},{"doc":"defpacket Macro First, we'll take a look at how the ElvenGard.Network.PacketSerializer.defpacket/3 macro works. Here is the most basic example for a packet definition that you can have: # Textual packet defpacket &quot;HEADER&quot; , as : TextualPacket # Binary packet defpacket 0x0000 , as : BinaryPacket This will generate a simple structure, without any field. The first parameter of the defpacket is its ID. It's used to identify it, so that we know which fields to decode next. We also have to specify a :as option. This is the name of the generated structure. For example, the following code will generate a structure called MyApp.MyPacket with the binary packet header 0x0000 . defmodule MyApp do defpacket 0x0000 , as : MyPacket end Guards Now, let's imagine you have several packets with the same packet header and to find out how to deserialize it, you need to access a socket-related state. To do this, you can use socket assigns and add a guard to the packet: defpacket 0x0000 when socket . assigns . state == :init , as : InitPacket defpacket 0x0000 when socket . assigns . state == :sync , as : SyncPacket In this example, if when deserializing the packet, the socket state is :init , the structure returned by our deserializer will be InitPacket and if it is :sync , it will be SyncPacket . For more information and examples of how to use guards, I recommend you read the code in examples/minecraft_ex . NOTE : guards are only available for deserialization (client packets). Body Finally, to define fields in our packet, their name, type etc., we'll use a do ... end block. defpacket 0x0000 , as : MyPacket do field ... end field Macro Now that you know how to create a packet, it's time to see how to create fields. To do this, you can use the field macro. It's very easy to use: field :field1 , FieldType field :field2 , FieldType , opt : :value Here, the first line will add a field to our packet with field1 as name and FieldType as type. The name must be an atom and the field type a module which use the ElvenGard.Network.Type behaviour. The third parameter is optional and is a keyword list defining options to send when types will be decoded (see ElvenGard.Network.Type.encode/2 and ElvenGard.Network.Type.decode/2 ). Decorators Using defpacket and field , you can now create structures with different fields, but you still can't serialize or deserialize them in order to send them to or receive them from the client. ElvenGard.Network defined 2 decorators for this: @serializable true : indicates that the packet is intended to be serialized. It is therefore a server packet . @deserializable true : indicates that the packet is intended to be deserialized. It is therefore a client packet . By tagging a packet as deserializable , the defpacket macro will automatically create a deserialize/1 function inside the generated module. Tagging a packet as deserializable/3 will, this time, create 2 deserialize/3 functions, one inside the generated module and one outside it. The latter will redirect to the former. These helper functions will come in very handy in our Network Codec , as we'll see in the next chapter. Now that we know the theory, it's time to put it into practice.","ref":"packet_definitions.html#macros","title":"Packet Definitions - Macros","type":"extras"},{"doc":"Let's first define our client packets. # file: lib/login_server/client_packets.ex defmodule LoginServer.ClientPackets do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.ClientPackets &quot;&quot;&quot; use ElvenGard.Network.PacketSerializer alias LoginServer.Types.StringType ## Ping packet @deserializable true defpacket &quot;PING&quot; , as : PingRequest ## Login packet @deserializable true defpacket &quot;LOGIN&quot; , as : LoginRequest do field :username , StringType field :password , StringType end end These packets are pretty straightforward, so there's not much to explain. Just note the presence of @deserializable true , which clearly indicates that these packets are intended to be deserialized as they are received from the client.","ref":"packet_definitions.html#client-packets","title":"Packet Definitions - Client Packets","type":"extras"},{"doc":"Now let's define the server's one. # file: lib/login_server/server_packets.ex defmodule LoginServer.ServerPackets do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.ServerPackets &quot;&quot;&quot; use ElvenGard.Network.PacketSerializer alias LoginServer.Types . { DateTimeType , StringType } alias LoginServer.SubPackets.WorldInfo ## Ping packet @serializable true defpacket &quot;PONG&quot; , as : PongResponse do field :time , DateTimeType end ## Login packets @serializable true defpacket &quot;FAIL&quot; , as : LoginFailed do field :reason , StringType end @serializable true defpacket &quot;SUCCESS&quot; , as : LoginSucceed do field :world , WorldInfo , sep : &quot;:&quot; end end Once again, these packets are pretty straightforward. We use @serializable true to indicate that we will send them to the client. Moreover, we're using the WorldInfo sub-packet which, remember, expects the field separator as an option. This is where we define it.","ref":"packet_definitions.html#server-packets","title":"Packet Definitions - Server Packets","type":"extras"},{"doc":"You now know how to create a packet that can be serialized and deserialized. Now it's time to encode or decode it so that it can transit over the network.","ref":"packet_definitions.html#summary","title":"Packet Definitions - Summary","type":"extras"},{"doc":"In this section, we will learn how to use ElvenGard.Network.NetworkCodec . A NetworkCodec is a behaviour that define how a packet should transit over the network. It must define 3 callbacks : next/2 : this callback takes a raw binary and a socket as parameters. It returns the first packet found in the form of a tuple: {packet_binary, rest_of_binary} . decode/2 : this callback takes the binary return by next/2 and returns the deserialized packet (a structure). You should use the deserialize/3 helper created by defpacket encode/2 : this callback will be called when you call ElvenGard.Network.Socket.send/2 with your module as encoder. It returns an iodata that will, then, be send to your client. You should use the serialize/1 helper created by defpacket","ref":"network_codec.html","title":"Network Codec","type":"extras"},{"doc":"# file: lib/login_server/endpoint/network_codec.ex defmodule LoginServer.Endpoint.NetworkCodec do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.Endpoint.NetworkCodec &quot;&quot;&quot; @behaviour ElvenGard.Network.NetworkCodec alias LoginServer.ClientPackets ## Behaviour impls @impl true def next ( raw , _socket ) do case String . split ( raw , &quot; \\n &quot; , parts : 2 ) do [ packet ] -&gt; { packet , &quot;&quot; } [ packet , rest ] -&gt; { packet , rest } end end @impl true def decode ( raw , socket ) do case String . split ( raw , &quot; &quot; , parts : 2 ) do [ packet_id ] -&gt; ClientPackets . deserialize ( packet_id , &quot;&quot; , socket ) [ packet_id , params ] -&gt; ClientPackets . deserialize ( packet_id , params , socket ) end end @impl true def encode ( struct , _socket ) when is_struct ( struct ) do # %LoginSucceed{world_info: %WorldInfo{host: &quot;127.0.0.1&quot;, port: 5000}} struct # {&quot;SUCCESS&quot;, [&quot;127.0.0.1&quot;, &quot;:&quot;, &quot;5000&quot;]} |&gt; struct . __struct__ . serialize ( ) # [&quot;SUCCESS&quot;, [&quot;127.0.0.1&quot;, &quot;:&quot;, &quot;5000&quot;]] |&gt; Tuple . to_list ( ) # [&quot;SUCCESS&quot;, &quot; &quot;, [&quot;127.0.0.1&quot;, &quot;:&quot;, &quot;5000&quot;]] |&gt; Enum . intersperse ( &quot; &quot; ) # [[&quot;SUCCESS&quot;, &quot; &quot;, [&quot;127.0.0.1&quot;, &quot;:&quot;, &quot;5000&quot;]], &quot;\\n&quot;] |&gt; then ( &amp; [ &amp;1 | &quot; \\n &quot; ] ) # This iolist will be send as `SUCCESS 127.0.0.1:5000\\n` end end Here, our decode/2 callback uses LoginServer.ClientPackets.deserialize/3 , which redirects to LoginServer.ClientPackets.&lt;YOUR_PACKET_STRUCTURE&gt;.deserialize/3 . Note also that encode/2 uses LoginServer.ServerPackets.&lt;YOUR_PACKET_STRUCTURE&gt;.serialize/1 to serialize our packet. Finally, we add a \\n at the end of our packet as defined by the network protocol .","ref":"network_codec.html#create-a-network-codec","title":"Network Codec - Create a Network Codec","type":"extras"},{"doc":"Now that we've defined our functions for transforming data structures, it's time to see how to link them to our Protocol .","ref":"network_codec.html#summary","title":"Network Codec - Summary","type":"extras"},{"doc":"In this section, we will learn how to use ElvenGard.Network.View . The role of a View is to prepare data to be sent via a packet structure. For example, we may have a packet taking a string as a field, but our backend returns a map. Encoding our map every time we need to create a packet would be problematic, as it would create code duplication and would be a potential source of errors. Other common uses are, for example, when a packet has one or more sub-packets, or when we need to set default values for certain fields. For simplicity's sake, and to avoid having to duplicate this logic in every function that needs to render packets, views have been created.","ref":"packet_views.html","title":"Packet Views","type":"extras"},{"doc":"# file: lib/login_server/packet_views.ex defmodule LoginServer.PacketViews do @moduledoc &quot;&quot;&quot; Documentation for LoginServer.PacketViews &quot;&quot;&quot; use ElvenGard.Network.View alias LoginServer.ServerPackets . { PongResponse , LoginFailed , LoginSucceed } alias LoginServer.SubPackets.WorldInfo @impl true def render ( :pong_response , %{ time : time } ) do % PongResponse { time : time } end def render ( :login_failed , %{ reason : reason } ) do % LoginFailed { reason : reason } end def render ( :login_succeed , %{ world : world } ) do % LoginSucceed { world : % WorldInfo { host : world . host , port : world . port } } end end Not much to explain, except that the render/2 callback takes as its first parameter an identifier (which must be a String or an Atom ) and as its second parameter attributes enabling us to generate our packet. Most of the time, the callback returns a strucure (a packet), but in exceptional cases it can also directly return iodata.","ref":"packet_views.html#create-a-packet-views","title":"Packet Views - Create a Packet Views","type":"extras"},{"doc":"You now know how to create Views: functions for creating server packets. In the final chapter, we'll use this module to create and send packets to our client.","ref":"packet_views.html#summary","title":"Packet Views - Summary","type":"extras"},{"doc":"If you remember, in our config.exs , we defined the following lines: config :login_server , LoginServer.Endpoint.Protocol , packet_handler : LoginServer.Endpoint.PacketHandler , ... Once the packets have been decoded by our Network Codec, they will be redirected to this module. This module must implement the ElvenGard.Network.PacketHandler protocol. A Packet Handler is the module that will manage the logic associated with our client packets. Let's see how to create one for our demo application.","ref":"packet_handler.html","title":"Packet Handler","type":"extras"},{"doc":"# file: lib/login_server/endpoint/packet_handler.ex defmodule LoginServer.Endpoint.PacketHandler do @moduledoc &quot;&quot;&quot; LoginServer.Endpoint.PacketHandler &quot;&quot;&quot; @behaviour ElvenGard.Network.PacketHandler alias ElvenGard.Network.Socket alias LoginServer.ClientPackets . { PingRequest , LoginRequest } alias LoginServer.PacketViews ## Handlers @impl true def handle_packet ( % PingRequest { } , socket ) do render = PacketViews . render ( :pong_response , %{ time : DateTime . utc_now ( ) } ) :ok = Socket . send ( socket , render ) { :cont , socket } end def handle_packet ( % LoginRequest { username : username , password : password } , socket ) do render = if auth_using_db ( username , password ) do PacketViews . render ( :login_succeed , %{ world : get_worlds_from_manager ( ) } ) else PacketViews . render ( :login_failed , %{ reason : &quot;Bad credentials :/`&quot; } ) end :ok = Socket . send ( socket , render ) { :halt , socket } end ## Fake functions defp auth_using_db ( username , password ) do case { username , password } do { &quot;admin&quot; , &quot;password&quot; } -&gt; true _ -&gt; false end end defp get_worlds_from_manager ( ) do %{ host : &quot;127.0.0.1&quot; , port : 5000 } end end Note that each handler takes as parameter a structure with decoded fields representing the packet and the socket associated with it. A handler must return {:cont, new_socket} if we want to continue receiving packets, or {:halt, new_socket} to close the connection to the socket and shutdown the associated GenServer. Here, our handlers are quite simple: the first will just send to our client a PongResponse packet with the current time and the second will return a LoginSucceed or a LoginFailed depending on the credentials passed in parameter. Note also that after a PingRequest , we'll continue to handle packets, whereas after a LoginRequest we'll automatically close the connection.","ref":"packet_handler.html#create-a-packethandler","title":"Packet Handler - Create a PacketHandler","type":"extras"},{"doc":"In this section, we've learned how to handle client packets, create logic around them and use the previously created Views . Our application is now ready for use. You can view the whole source code at examples/login_server .","ref":"packet_handler.html#summary","title":"Packet Handler - Summary","type":"extras"}]